//
//	W2OController.m
//	W2O
//
//	Created by Woon Seung Yeo on 2/26/06.
//	Based on Apple's "RFCOMM_Open_SPP_Example" sample codes
//
//	Copyright 2006 CCRMA, Stanford University. All rights reserved.
//

#import "W2OController.h"

#define max(a,b)	((a>b)?(a):(b))
#define min(a,b)	((a<b)?(a):(b))

#define RADIAN_TO_DEGREE	180/M_PI

// Bluetooth connection patameters: specific to WiTilt v.2
#define BAUD_RATE	57600
#define DATA_BITS	8
#define PARITY		kBluetoothRFCOMMParityTypeNoParity
#define STOP_BITS	1

// OSC client parameters
#define IPADDRESS	@"127.0.0.1"
#define PORTNUMBER	1228

#define ADDR_RAW	@"/w2o/raw"
#define ADDR_DEG	@"/w2o/deg"
#define ADDR_MAG	@"/w2o/mag"

#define TYPETAG_RAW	"iiii"
#define TYPETAG_DEG	"ifff"
#define TYPETAG_MAG	"if"

@implementation W2OController

- (void)readCalibrationData {
	// Calibration data file path: "Resource" path
	NSString *calibrationDataFilePath = 
		[[NSBundle mainBundle] pathForResource:@"Calibration" ofType:@"txt"];

	// Read calibration data from the path found above.
	NSString *calibrationDataString = 
		[NSString stringWithContentsOfFile:calibrationDataFilePath encoding:NSUTF8StringEncoding error:NULL];

	// Divide string by " "
	NSArray *calibrationDataArray = [calibrationDataString componentsSeparatedByString:@" "];

	int i;
	for (i=0;i<3;i++) {
		_calibrationDataMaxima[i] = [[calibrationDataArray objectAtIndex:i*2] floatValue];
		[[_calibrationDataMaxForm cellAtIndex:i] setIntValue:_calibrationDataMaxima[i]];

		_calibrationDataMinima[i] = [[calibrationDataArray objectAtIndex:i*2+1] floatValue];
		[[_calibrationDataMinForm cellAtIndex:i] setIntValue:_calibrationDataMinima[i]];

		_calibrationDataMean[i] = (_calibrationDataMaxima[i] + _calibrationDataMinima[i]) / 2;
		_calibrationDataScalingAmp[i] = (_calibrationDataMaxima[i] - _calibrationDataMinima[i]) / 2;
	}
	// _g = THE g!
	_g = _calibrationDataMaxima[2] - _calibrationDataMean[2];

	return;
}

- (void)saveCalibrationData {
	// Format an NSString to update calibration data
	NSString *calibrationDataString = 
		[NSString stringWithFormat:@"%3.1f %3.1f %3.1f %3.1f %3.1f %3.1f",
			_calibrationDataMaxima[0], _calibrationDataMinima[0],
			_calibrationDataMaxima[1], _calibrationDataMinima[1],
			_calibrationDataMaxima[2], _calibrationDataMinima[2]];
	
	// Calibration data file path: "Resource" path
	NSString *calibrationDataFilePath = 
		[[NSBundle mainBundle] pathForResource:@"Calibration" ofType:@"txt"];

	// Save calibration data to the path found above.
	[calibrationDataString writeToFile:calibrationDataFilePath atomically:YES encoding:NSUTF8StringEncoding error:NULL];
}

- (void)awakeFromNib {
	[_IPField setStringValue:IPADDRESS];
	[_portField setIntValue:PORTNUMBER];

	[_tiltAccelRawAddressField setStringValue:ADDR_RAW];
	[_tiltAccelDegAddressField setStringValue:ADDR_DEG];
	[_accelerationMagAddressField setStringValue:ADDR_MAG];
	
	[self readCalibrationData];
	
	[_openButton setEnabled:YES];
}

- (IBAction)closeConnectonAction:(id)sender {
	// Do the real work to close the connection:
	[self closeRFCOMMConnectionOnChannel:mRFCOMMChannel];

	// The button did its job until we open a new connection we do not need to re-enable it:
	[_openButton setEnabled:YES];
	[_startButton setEnabled:NO];
	[_stopButton setEnabled:NO];
	[_closeButton setEnabled:NO];

	[_deviceNameField setStringValue:@""];
	[_macAddressField setStringValue:@""];
}

- (IBAction)openConnectionAction:(id)sender {
	if ( [self openSerialPortProfile] )
	{
		// if openSerialPortProfile is successful the connection is open or at
		// least in the process of opening. So we disable the "Open" button. The
		// button will be re-enabled if the open process fails or when the
		// connection is closed.
		[_openButton setEnabled:NO];
		[_startButton setEnabled:YES];
		[_stopButton setEnabled:NO];
		[_closeButton setEnabled:NO];
	}
}

- (IBAction)startDataStream:(id)sender {
	if (mRFCOMMChannel) {
		if (_oscPort) [_oscPort release];
		_oscPort = [OSCPort oscPortToAddress:(char *)[[_IPField stringValue] UTF8String] portNumber:[_portField intValue]];
		[_oscPort retain];

		[mRFCOMMChannel writeSync:"1" length:1];
		[mRFCOMMChannel writeSync:"R" length:1];
		[mRFCOMMChannel writeSync:"S" length:1];

		[_IPField setEnabled:NO];
		[_portField setEnabled:NO];

		[_openButton setEnabled:NO];
		[_startButton setEnabled:NO];
		[_stopButton setEnabled:YES];
		[_closeButton setEnabled:NO];
	}
}

- (IBAction)stopDataStream:(id)sender {
	if (mRFCOMMChannel) {
		[mRFCOMMChannel writeSync:"A" length:1];
		
		[_IPField setEnabled:YES];
		[_portField setEnabled:YES];

		[_openButton setEnabled:NO];
		[_startButton setEnabled:YES];
		[_closeButton setEnabled:YES];
		[_stopButton setEnabled:NO];
	}
}

- (BOOL)windowShouldClose:(id)sender
{
	// If we have an open connection close:
	if ( ( mRFCOMMChannel != nil ) || ( mBluetoothDevice != nil ) )
		[self closeRFCOMMConnectionOnChannel:mRFCOMMChannel];

	return TRUE;
}

- (IBAction)setCalibrationMaxValue:(id)sender {
	int i;
	for (i=0;i<3;i++) {
		_calibrationDataMaxima[i] = [[sender cellAtIndex:i] intValue];
		[self saveCalibrationData];
	}
}

- (IBAction)setCalibrationMinValue:(id)sender {
	int i;
	for (i=0;i<3;i++) {
		_calibrationDataMinima[i] = [[sender cellAtIndex:i] intValue];
		[self saveCalibrationData];
	}
}

// =============================
// == BLUETOOTH SPECIFIC CODE ==
// =============================

- (BOOL)openSerialPortProfile
{
    IOBluetoothDeviceSelectorController	*deviceSelector;
	IOBluetoothSDPUUID					*sppServiceUUID;
	NSArray								*deviceArray;
	
    // The device selector will provide UI to the end user to find a remote device
    deviceSelector = [IOBluetoothDeviceSelectorController deviceSelector];
	if ( deviceSelector == nil )
	{
		NSLog( @"Error - unable to allocate IOBluetoothDeviceSelectorController.\n" );
		return FALSE;
	}
		
	// Create an IOBluetoothSDPUUID object for the chat service UUID
	sppServiceUUID = [IOBluetoothSDPUUID uuid16:kBluetoothSDPUUID16ServiceClassSerialPort];

	// Tell the device selector what service we are interested in.
	// It will only allow the user to select devices that have that service.
	[deviceSelector addAllowedUUID:sppServiceUUID];
	
	// Run the device selector modal.  This won't return until the user has selected a device and the device has
	// been validated to contain the specified service or the user has hit the cancel button.
	if ( [deviceSelector runModal] != kIOBluetoothUISuccess )
	{
		NSLog( @"User has cancelled the device selection.\n" );
		return FALSE;
	}

	// Get the list of devices the user has selected.
	// By default, only one device is allowed to be selected.
	deviceArray = [deviceSelector getResults];
	
	if ( ( deviceArray == nil ) || ( [deviceArray count] == 0 ) )
	{
		NSLog( @"Error - no selected device.  ***This should never happen.***\n" );
		return FALSE;
	}
	
	// The device we want is the first in the array (even if the user somehow selected more than
	// one device in this example we care only about the first one):
	IOBluetoothDevice *device = [deviceArray objectAtIndex:0];
	[_deviceNameField setStringValue:[device getName]];
	[_macAddressField setStringValue:[device getAddressString]];
	
	// Finds the service record that describes the service (UUID) we are looking for:
	IOBluetoothSDPServiceRecord	*sppServiceRecord = [device getServiceRecordForUUID:sppServiceUUID];
	
	if ( sppServiceRecord == nil )
	{
		NSLog( @"Error - no spp service in selected device.  ***This should never happen since the selector forces the user to select only devices with spp.***\n" );
		return FALSE;
	}

	// To connect we need a device to connect and an RFCOMM channel ID to open on the device:
	UInt8	rfcommChannelID;
	if ( [sppServiceRecord getRFCOMMChannelID:&rfcommChannelID] != kIOReturnSuccess )
	{
		NSLog( @"Error - no spp service in selected device.  ***This should never happen an spp service must have an rfcomm channel id.***\n" );
		return FALSE;
	}

	// Open asyncronously the rfcomm channel when all the open sequence is completed my implementation of "rfcommChannelOpenComplete:" will be called.
	if ( ( [device openRFCOMMChannelAsync:&mRFCOMMChannel withChannelID:rfcommChannelID delegate:self] != kIOReturnSuccess ) && ( mRFCOMMChannel != nil ) )
	{
		// Something went bad (looking at the error codes I can also say what, but for the moment let's not dwell on
		// those details). If the device connection is left open close it and return an error:
		NSLog( @"Error - open sequence failed.***\n" );
		
		[self closeDeviceConnectionOnDevice:device];
		
		return FALSE;
	}

	// So far a lot of stuff went well, so we can assume that the device is a good one and that rfcomm channel open process is going
	// well. So we keep track of the device and we (MUST) retain the RFCOMM channel:
	mBluetoothDevice = device;
	[mBluetoothDevice  retain];
	[mRFCOMMChannel retain];
	
	[mRFCOMMChannel setSerialParameters:BAUD_RATE
					dataBits:DATA_BITS 
					parity:PARITY
					stopBits:STOP_BITS];

	return TRUE;
}

- (void)closeRFCOMMConnectionOnChannel:(IOBluetoothRFCOMMChannel*)channel
{
	if ( mRFCOMMChannel == channel )
	{
		[mRFCOMMChannel closeChannel];
	}
}

- (void)closeDeviceConnectionOnDevice:(IOBluetoothDevice*)device
{
	if ( mBluetoothDevice == device )
	{
		IOReturn error = [mBluetoothDevice closeConnection];
		if ( error != kIOReturnSuccess )
		{
			// I failed to close the connection, maybe the device is busy, no problem, as soon as the device is no more busy it will close the connetion itself.
			NSLog(@"Error - failed to close the device connection with error %08lx.\n", (UInt32)error);
		}
		
		[mBluetoothDevice release];
		mBluetoothDevice = nil;
	}

	// Re-enable the open button so the user can restart the sequence:
	[_openButton setEnabled:YES];	

	[_startButton setEnabled:NO];
	[_stopButton setEnabled:NO];

	// Since we are closed also disables the close button:
	[_closeButton setEnabled:NO];
}


// Called by the RFCOMM channel on us once the baseband and rfcomm connection is completed:
- (void)rfcommChannelOpenComplete:(IOBluetoothRFCOMMChannel*)rfcommChannel status:(IOReturn)error
{
	// If it failed to open the channel call our close routine and from there the code will
	// perform all the necessary cleanup:
	if ( error != kIOReturnSuccess )
	{
		NSLog(@"Error - failed to open the RFCOMM channel with error %08lx.\n", (UInt32)error);
		[self rfcommChannelClosed:rfcommChannel];
		return;
	}

	// Now that the channel is successfully open we enable the close button:
	//[_openButton setEnabled:YES];

	// The RFCOMM channel is now completly open so it is possible to send and receive data
	// ... add the code that begin the send data ... for example to reset a modem:
	[rfcommChannel writeSync:"ATZ\n" length:4];
}

// Called by the RFCOMM channel on us when new data is received from the channel:
- (void)rfcommChannelData:(IOBluetoothRFCOMMChannel *)rfcommChannel data:(void *)dataPointer length:(size_t)dataLength
{
	short *index;
	short rawDataBuffer[3];
	float unbiasedBuffer[3];
	float angleBuffer[3];

	char *dataPtr = (char *)dataPointer;
	if (strncmp(dataPointer, "#@", 2) == 0) {
		while (dataLength >= 11) {
			// Index
			dataPtr = dataPtr + 2;
			index = (short *)dataPtr;

			// X, Y, and Z
			int i;
			for (i=0;i<3;i++) {
				dataPtr = dataPtr + 2;
				rawDataBuffer[i] = *(short *)dataPtr;
				unbiasedBuffer[i] = (float)rawDataBuffer[i] - _calibrationDataMean[i];
			}
			dataPtr = dataPtr + 3;

			// Send out OSC messages based on output configuration
			// Raw data
			if ([_tiltAccelRawButton state] == NSOnState)
				[_oscPort sendTo:(char *)[[_tiltAccelRawAddressField stringValue] UTF8String] 
					types:TYPETAG_RAW, *index, rawDataBuffer[0], rawDataBuffer[1], rawDataBuffer[2]];
			// Angle (degree)
			if ([_tiltAccelDegButton state] == NSOnState) {
				for (i=0;i<3;i++) {
					angleBuffer[i] = 
						RADIAN_TO_DEGREE * asin( max( min(unbiasedBuffer[i]/_calibrationDataScalingAmp[i], 1.0), -1.0) );
				}
				[_oscPort sendTo:(char *)[[_tiltAccelDegAddressField stringValue] UTF8String] 
					types:TYPETAG_DEG, *index, angleBuffer[0], angleBuffer[1], angleBuffer[2]];
			}
			// Accelerationn (magnitude)
			if ([_accelerationMagButton state] == NSOnState) {
				float magnitude = 
					(sqrt( pow(unbiasedBuffer[0],2)+pow(unbiasedBuffer[1],2)+pow(unbiasedBuffer[2],2) ) - _g) / _g * 100.0;
				[_oscPort sendTo:(char *)[[_accelerationMagAddressField stringValue] UTF8String]
					types:TYPETAG_MAG, *index, max(magnitude,0)];
			}
			dataLength = dataLength - 11;
		}
	}
}

// Called by the RFCOMM channel on us when something happens and the connection is lost:
- (void)rfcommChannelClosed:(IOBluetoothRFCOMMChannel *)rfcommChannel
{
	// wait a second and close the device connection as well:
	[self performSelector:@selector(closeDeviceConnectionOnDevice:) withObject:mBluetoothDevice afterDelay:1.0];
}

@end
